/*

 UNLu
 Departamento de Ciencias Basicas
 Division Estadistica y Sistemas
 Licenciatura en Sistemas de Informacion
 11411 Base de Datos
 Lic. Guillermo Cherencio

 Trabajo Practico I "Listado Especifico - Listado Generico"
 Ejercicio I, II, III
 Usando compilador GNU C/C++ GCC for Windows (cygwin)
 Libreria de funciones para manejo de archivos relativos
 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "librelativo.h"

int rel_init(rel_fd *f) { // inicializo archivo relativo
   int ret=0;
   if ( f->filename == NULL || *f->filename == '\0' ) {
      fprintf(stderr,"rel_init():Error Nombre de Archivo equivocado\n");
   } else if (f->reglen <= 0) {
      fprintf(stderr,"rel_init():Error Largo de Registro equivocado\n");
   } else if (f->maxreg <= 0) {
      fprintf(stderr,"rel_init():Error Maximo numero de registros equivocado\n");
   } else {
      if ( f->fp != NULL ) { // archivo abierto previamente
         fclose(f->fp);
         f->fp = NULL;
      }
      if ( (ret=rel_nuevobuffer(f)) ) {
         f->fp = fopen(f->filename,"r");
         if (f->fp == NULL) { // crear archivo
            ret = rel_crear(f);
         } else { // verificar tamao
            int cr = rel_cantidadreg(f);
            if ( cr != f->maxreg ) {
               fprintf(stderr,"rel_init():Error de tamao en archivo %s %d registros vs %d registros\n",f->filename,
                       cr,f->maxreg);
               fclose(f->fp);
               f->fp = NULL;
               ret = rel_crear(f);
            } else {
               fclose(f->fp);
               f->fp = fopen(f->filename,"r+");
               ret=1;
            }
         }
      } else fprintf(stderr,"rel_init():Error en creacion de nuevo buffer, archivo %s no inicializado\n",f->filename);
   }
   return ret;
}

int rel_finit(rel_fd *f) {  // dejo de usar archivo relativo
   if ( f == NULL ) return 0;
   if ( f->fp != NULL ) { // archivo abierto previamente
      fclose(f->fp);
      f->fp = NULL;
   }
   if ( f->buffer != NULL ) free(f->buffer); // buffer inicializado previamente
   if ( f->filename != NULL ) free(f->filename);
   free(f);
   return 1;
}

long rel_tamanio(const rel_fd *f) {  // largo absoluto en bytes de archivo
   long opos = ftell(f->fp),pos=0L;
   if ( fseek(f->fp,0L,SEEK_END) == 0 ) pos = ftell(f->fp);
   if ( opos != pos ) fseek(f->fp,opos,SEEK_SET);
   return pos;
}

int rel_cantidadreg(const rel_fd *f) {  // cantidad total de registros 
   return (int) (rel_tamanio(f) / f->reglen);
}

int rel_limpiar(rel_fd *f) {  // inicializa buffer con \0
   if ( f->buffer == NULL ) rel_nuevobuffer(f);
   if ( f->buffer != NULL ) {
      memset(f->buffer,'\0',f->reglen+1);
      return 1;
   } else {
      fprintf(stderr,"rel_limpiar():Error buffer no inicializado\n");
      return 0;
   }
}

int rel_nuevobuffer(rel_fd *f) {  // crea nuevo buffer inicializado con \0
   if ( f->buffer != NULL ) free(f->buffer); // buffer inicializado previamente
   f->buffer = (char *) malloc(f->reglen+1);
   if ( f->buffer == NULL ) {
      fprintf(stderr,"rel_nuevobuffer():Error en asignacion de memoria\n");
      return 0;
   } else return rel_limpiar(f);
}

int rel_crear(rel_fd *f) {  // crea nuevo archivo y lo inicializa con registros que contienen \0
   int reg=0;
   if ( f->fp == NULL ) f->fp = fopen(f->filename,"w");
   if ( f->fp != NULL ) {
      if (rel_limpiar(f)) {
         while ( rel_grabar(f,reg) ) reg++;
         fclose(f->fp);
         f->fp = fopen(f->filename,"r+");
         return 1;
      }
   }
   return 0;
}

int rel_leer(const rel_fd *f,const int reg) {  // lee el registro indicado y lo carga en buffer
   if (!(reg >= 0 && reg < f->maxreg)) return 0;
   if ( rel_apuntar(f,reg) ) {
      if ( fread(f->buffer,1,f->reglen,f->fp) == f->reglen) {
         return 1;
      } else {
         fprintf(stderr,"rel_leer():Error al intentar leer registro %d en %s\n",reg,f->filename);
      }
   }
   return 0;
}

int rel_grabar(const rel_fd *f,const int reg) {  // graba el registro indicado desde buffer
   if (!(reg >= 0 && reg < f->maxreg)) return 0;
   if ( rel_apuntar(f,reg) ) {
      if ( fwrite(f->buffer,1,f->reglen,f->fp) == f->reglen ) {
         return 1;
      } else {
         fprintf(stderr,"rel_grabar():Error al intentar grabar registro %d en %s\n",reg,f->filename);
      }
   } 
   return 0;
}

int rel_apuntar(const rel_fd *f,const int reg) {  // apunta archivo a la posicion absoluta del registro indicado
   long pos = (long) reg * f->reglen;
   if ( ftell(f->fp) != pos ) {
      if ( fseek(f->fp,pos,SEEK_SET) == 0 ) {
         return 1;
      } else {
         fprintf(stderr,"rel_apuntar():Error al intentar posicionarse en reg %d de archivo %s\n",reg,f->filename);
         return 0;
      }
   } else return 1;
}

rel_fd *rel_decl(const char *filename,const int maxreg,const int reglen) {
  rel_fd *fd = (rel_fd *) malloc(sizeof(rel_fd));
  if ( fd == NULL ) { fprintf(stderr,"rec_decl():Error en asignacion de memoria(1)\n");return NULL; }
  if ( filename != NULL && *filename != '\0') {
     fd->filename = (char *) malloc(strlen(filename)+1);
     if ( fd->filename != NULL ) strcpy(fd->filename,filename);
     else fprintf(stderr,"rel_decl():Error en asignacion de memoria(2)\n");
  } else {
     fprintf(stderr,"rel_decl():Error en nombre de archivo\n");
  }
  fd->reglen   = reglen;
  fd->maxreg   = maxreg;
  fd->fp       = NULL;
  fd->buffer   = NULL;
//printf("rel_decl(): |%s|%d|%d|\n",fd->filename,fd->maxreg,fd->reglen);
  return fd;
}

/************************************************************************************
  FIN LIBRERIA
************************************************************************************/
