/*

 UNLu
 Departamento de Ciencias Basicas
 Division Estadistica y Sistemas
 Licenciatura en Sistemas de Informacion
 11411 Base de Datos
 Lic. Guillermo Cherencio

 Trabajo Practico I "Listado Especifico - Listado Generico"
 Ejercicio I
 Implementacion de archivo relativo usando compilador GNU C/C++ GCC for Windows (cygwin)

 =============script de compilacion, linkedicion y ejecucion=============================
#!/bin/sh
#
# script para compilar aplicacion bdtp11.c 
#
#                Atte. Guillermo Cherencio
#                UNLu
#
echo compilo - enlazo - ejecuto
gcc -Wall -g bdtp11.c libtp1.c -o ../../bin/bdtp11.exe -I../../include -L../../lib -lmilib01
../../bin/bdtp11.exe
 ==========================================================================================

 =============estructura de directorios de trabajo=========================================
C:.
+---bin
+---include
+---lib
+---obj
|   \---lib1
\---src
    +---bdtp1
    +---lib1
    \---prj1
 ==========================================================================================
 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libtp1.h"

/* parametros */

#define LIBRE   '0'
#define OCUPADO '1'
#define ARCHIVO "datos.dat"
#define MAXREG 1000
#define REGLEN 65 // 1 + 30 + 30 + 4  tamao registro en disco

/* estructura de registro */

// tamao de registro en memoria: el sizeof(rg) es 68

typedef struct {
   char est;
   char ape[30];
   char nom[30];
   int cod;
} rg;

/* prototipo de funciones */
void crear();
void ingresar();
void borrar();
void listar();

void mostrar(rg *);
void tipear(rg *);
int validar(rg *);
void rellenar(rg *);
void limpiar(rg *);
void parseBuf(char *,rg *);
void parseReg(rg *,char *);
int leer(FILE *fp,int pos,rg *reg);
int grabar(FILE *fp,int pos,rg *reg);
int apuntar(FILE *fp,long pos);


/************************************************************************************
  COMIENZO PROGRAMA
************************************************************************************/
int main() {
  int op;
  char *titulo = "Listado Especifico - Generico (1)";
  // opcion "" indica que se termina el arreglo de opciones de menu
  char *opc[] = { "Crear Archivo","Altas y Modificaciones","Bajas","Listar", "" };
  /****************************************************************************************************
    para evitar el trabajo de hacer switch y preguntar por cada opcion elegida para ejecutar funcion
    creo arreglo con puntero a funciones y se lo paso a menu() para que realice las ejecuciones
    segun lo indicado por el usuario
  ****************************************************************************************************/
  void ((*pfuncion[]) ()) = { crear, ingresar, borrar, listar };
  do {
     op=menu(titulo,opc,pfuncion);
  } while(op != 99);
  return 0;
}
/************************************************************************************
  FIN PROGRAMA
************************************************************************************/

/* funciones alto nivel */

void crear() {
   if ( !preguntarSN("Genera un Archivo nuevo?\nLos Datos Actuales se perderan.")) return;
   FILE *fp = fopen(ARCHIVO,"w");
   if ( fp != NULL ) {
      int pos=0;
      rg reg;
      limpiar(&reg);
      while ( grabar(fp,pos,&reg) ) pos++;
      fclose(fp);
      printf("Archivo %s Generado Ok\n",ARCHIVO);
   } else {
      printf("%s no pudo ser creado\n",ARCHIVO);
   }
}

void ingresar() {
   FILE *fp = fopen(ARCHIVO,"r+");
   if ( fp != NULL ) {
      rg regm, regf;
      int pos;
      do {
          pos = getIntP("Posicion (-1 para salir): ");
          if ( pos != -1 ) {
             if ( pos >= 0 && pos < MAXREG ) {
                if ( leer(fp,pos,&regf) ) {
                   if ( regf.est == OCUPADO ) {
                      printf("Posicion ya ocupada!\n");
                      mostrar(&regf);
                   }
                   regm.est = OCUPADO;
                   tipear(&regm);
                   if ( validar(&regm) ) {
                      if (preguntarSN("Cambia esta Posicion?")) {
                         grabar(fp,pos,&regm);
                      }
                   }
                }
             } else {
                printf("%d es una posicion no valida (%d - %d)\n",pos,0,MAXREG);
             }
          }
      } while(pos != -1);
      fclose(fp);
   } else {
      printf("%s no pudo ser abierto\n",ARCHIVO);
   }
}

void borrar() {
   FILE *fp = fopen(ARCHIVO,"r+");
   if ( fp != NULL ) {
      rg regf;
      int pos;
      do {
          pos = getIntP("Posicion (-1 para salir): ");
          if ( pos != -1 ) {
             if ( pos >= 0 && pos < MAXREG ) {
                if ( leer(fp,pos,&regf) ) {
                   if ( regf.est == OCUPADO ) {
                      mostrar(&regf);
                      if (preguntarSN("Borra esta Posicion?")) {
                         regf.est = LIBRE;
                         grabar(fp,pos,&regf);
                      }
                   } else {
                      printf("%d es una posicion no ocupada\n",pos);
                   }
                }
             } else {
                printf("%d es una posicion no valida (%d - %d)\n",pos,0,MAXREG);
             }
          }
      } while(pos != -1);
      fclose(fp);
   } else {
      printf("%s no pudo ser abierto\n",ARCHIVO);
   }
}

void listar() {
   FILE *fp = fopen(ARCHIVO,"r");
   if ( fp != NULL ) {
      rg reg;
      int pos=0;
      printf("   # Cod.  Apellido                       Nombre\n");
      while ( leer(fp,pos,&reg) ) {
         if ( reg.est == OCUPADO ) {
            printf("%4d %4d %30s %30s\n",pos,reg.cod,reg.ape,reg.nom);
         }
         pos++;
      }
      fclose(fp);
   } else {
      printf("%s no pudo ser abierto\n",ARCHIVO);
   }
}

/* funciones de bajo nivel */

void mostrar(rg *reg) {
   printf("Codigo: %d\n",reg->cod);
   printf("Apellido: %s\n",reg->ape);
   printf("Nombre: %s\n",reg->nom);
}

void tipear(rg *reg) {
   reg->cod = getIntP("Codigo: ");
   strcpy(reg->ape,getStringP("Apellido: "));
   strcpy(reg->nom,getStringP("Nombre: "));
}

int validar(rg *reg) {
   int ret=1;
   if ( !(reg->cod > 0 && reg->cod <= 9999) ) { 
      printf("Codigo debe ser un numero entre %d y %d\n",1,9999);
      ret=0;
   } else if ( reg->ape == NULL ) {
      printf("Debe ingresar Apellido\n");
      ret=0;
   } else if ( reg->nom == NULL ) {
      printf("Debe ingresar Nombre\n");
      ret=0;
   }
   return ret;
}

void rellenar(rg *reg) {
   int n,largo1=sizeof(reg->ape),largo2=sizeof(reg->nom);
   for(n=strlen(reg->ape);n<largo1;n++) reg->ape[n] = ' ';
   for(n=strlen(reg->nom);n<largo2;n++) reg->nom[n] = ' ';
}

void limpiar(rg *reg) {
   int n,largo1=sizeof(reg->ape),largo2=sizeof(reg->nom);
   reg->est = LIBRE;
   reg->cod = -1;
   for(n=0;n<largo1;n++) reg->ape[n] = ' ';
   for(n=0;n<largo2;n++) reg->nom[n] = ' ';
}

int leer(FILE *fp,int pos,rg *reg) {
   char buf[REGLEN];
   if (!(pos >= 0 && pos < MAXREG)) return 0;
   if ( apuntar(fp,(long) pos*(REGLEN)) ) {
      if ( fread(buf,1,REGLEN,fp) == REGLEN) {
         parseReg(reg,buf);
         return 1;
      } else {
         printf("Error al intentar leer registro %d en %s\n",pos,ARCHIVO);
      }
   }
   return 0;
}

int grabar(FILE *fp,int pos,rg *reg) {
   char buf[REGLEN];
   if (!(pos >= 0 && pos < MAXREG)) return 0;
//printf("grabar esta en %d y debe ir a %d\n",ftell(fp),(long) (pos*REGLEN));
   if ( apuntar(fp,(long) pos*(REGLEN)) ) {
      rellenar(reg);
      parseBuf(buf,reg);
      if ( fwrite(buf,1,REGLEN,fp) == REGLEN ) {
         return 1;
      } else {
         printf("Error al intentar grabar registro %d en %s\n",pos,ARCHIVO);
      }
   } 
   return 0;
}

int apuntar(FILE *fp,long pos) {
   if ( ftell(fp) != pos ) {
      if ( fseek(fp,pos,SEEK_SET) == 0 ) {
         return 1;
      } else {
         printf("Error al intentar posicionarse en %s\n",ARCHIVO);
         return 0;
      }
   } else return 1;
}

void parseBuf(char *buf,rg *reg) {
   int i=0,w;
   char *p;
   buf[i++] = reg->est;
   for(w=0;w<sizeof(reg->ape);w++,i++) buf[i] = reg->ape[w];
   for(w=0;w<sizeof(reg->nom);w++,i++) buf[i] = reg->nom[w];
   p = buf+i;
   w = reg->cod;
   memcpy(p,&w,sizeof(int));
//   printf("pf(): ");for(i=0;i<REGLEN;i++) printf("%c",buf[i]);printf("\n");
}

void parseReg(rg *reg,char *buf) {
   int i=0,w;
   char *p;
   reg->est = buf[i++];
   for(w=0;w<sizeof(reg->ape);w++,i++) reg->ape[w] = buf[i];reg->ape[w-1]='\0';
   for(w=0;w<sizeof(reg->nom);w++,i++) reg->nom[w] = buf[i];reg->nom[w-1]='\0';
   p = buf+i;
   memcpy(&w,p,sizeof(int));
   reg->cod = w;
//   printf("pr(): %c %s %s %d\n",reg->est,reg->ape,reg->nom,reg->cod);
}


