/*

 UNLu
 Departamento de Ciencias Basicas
 Division Estadistica y Sistemas
 Licenciatura en Sistemas de Informacion
 11411 Base de Datos
 Lic. Guillermo Cherencio

 Trabajo Practico I "Listado Especifico - Listado Generico"
 Ejercicio II
 Implementacion de archivo relativo usando compilador GNU C/C++ GCC for Windows (cygwin)
 Abm sobre archivo relativo utilizando clave primaria codigo de cliente.
 Algoritmo de hashing: Modulo, dividiendo por numero primo 1009 (primo cercano a 1000),
 para mantener un factor de carga apropiado y utilizando como area de
 desborde los registros por encima de MAXREG (1009)
 Requerira de un proceso de reorganizacion del area de desborde, luego de un uso prolongado,
 ya que alli se pueden ir acumulando registros "dados de baja".
 A traves del campo ovr (desborde, overflow) permite implementar una lista simplemente 
 enlazada para el manejo de claves sinonimas

 =============script de compilacion, linkedicion y ejecucion=============================
#!/bin/sh
#
# script para compilar aplicacion bdtp11.c 
#
#                Atte. Guillermo Cherencio
#                UNLu
#
echo compilo - enlazo - ejecuto
gcc -Wall -g bdtp12.c libtp1.c -o ../../bin/bdtp12.exe -I../../include -L../../lib -lmilib01
../../bin/bdtp12.exe
 ==========================================================================================

 =============estructura de directorios de trabajo=========================================
C:.
+---bin
+---include
+---lib
+---obj
|   \---lib1
\---src
    +---bdtp1
    +---lib1
    \---prj1
 ==========================================================================================
 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libtp1.h"

/* parametros */

#define LIBRE   '0'
#define OCUPADO '1'
#define VACIO -1
#define ARCHIVO    "datos2.dat"
#define ARCHIVOTMP "datos.tmp"
#define MAXREG 1009
#define REGLEN 69 // 1 + 30 + 30 + 4 + 4 tamao registro en disco
#define MINPK 0 // valor minimo de pk
#define MAXPK 9999 // valor maximo de pk

/* estructura de registro */

// tamao de registro en memoria: el sizeof(rg) es 72

typedef struct {
   char est;
   char ape[30];
   char nom[30];
   int cod;
   int ovr;
} rg;

/* prototipo de funciones */
void crear();
void ingresar();
void borrar();
void listar();
void organizar();
void sinonimas();

void mostrar(rg *);
void tipear(rg *);
int validar(rg *);
void rellenar(rg *);
void limpiar(rg *);
void parseBuf(char *,rg *);
void parseReg(rg *,char *);
int leer(FILE *fp,int pos,rg *reg);
int grabar(FILE *fp,int pos,rg *reg);
int apuntar(FILE *fp,long pos);
int hashing(int pk);
int nuevoreg(FILE *fp);


/************************************************************************************
  COMIENZO PROGRAMA
************************************************************************************/
int main() {
  int op;
  char *titulo = "Listado Especifico - Generico (2)";
  // opcion "" indica que se termina el arreglo de opciones de menu
  char *opc[] = { "Crear Archivo","Altas y Modificaciones","Bajas","Listar", "ReOrganizar", "Busco Sinonimas", "" };
  /****************************************************************************************************
    para evitar el trabajo de hacer switch y preguntar por cada opcion elegida para ejecutar funcion
    creo arreglo con puntero a funciones y se lo paso a menu() para que realice las ejecuciones
    segun lo indicado por el usuario
  ****************************************************************************************************/
  void ((*pfuncion[]) ()) = { crear, ingresar, borrar, listar, organizar, sinonimas };
  do {
     op=menu(titulo,opc,pfuncion);
  } while(op != 99);
  return 0;
}
/************************************************************************************
  FIN PROGRAMA
************************************************************************************/

/* funciones alto nivel */

void crear() {
   if ( !preguntarSN("Genera un Archivo nuevo?\nLos Datos Actuales se perderan.")) return;
   FILE *fp = fopen(ARCHIVO,"w");
   if ( fp != NULL ) {
      int pos=0;
      rg reg;
      limpiar(&reg);
      while ( pos < MAXREG && grabar(fp,pos,&reg) ) pos++;
      fclose(fp);
      printf("Archivo %s Generado Ok\n",ARCHIVO);
   } else {
      printf("%s no pudo ser creado\n",ARCHIVO);
   }
}

void ingresar() {
   FILE *fp = fopen(ARCHIVO,"r+");
   if ( fp != NULL ) {
      rg regf;
      int dir,nvadir,pk;
      unsigned char encontre,error,limpio;
      do {
          pk = getIntP("Cod.Cliente (-1 para salir): ");
          if ( pk != -1 ) {
             if ( pk >= MINPK && pk <= MAXPK ) {
                dir = hashing(pk);
                nvadir = dir;
                encontre=0;error=0;limpio=0;
                while (!encontre && !error) {
                   if ( leer(fp,dir,&regf) ) {
                      if ( regf.est == OCUPADO ) {
                         if ( regf.cod == pk ) { // ya estaba dado de alta, regrabo
                            encontre=1;
                         } else { // maldicion! hay colision
                            if ( regf.ovr != VACIO ) { // ok, debo recorrer lista de overflow
                               dir = regf.ovr;
                            } else { // genero nuevo registro en area de desborde y lo enlazo
                               nvadir = nuevoreg(fp);
                               regf.ovr = nvadir;
                               if ( grabar(fp,dir,&regf) ) {
                                  dir = nvadir;  // apunto a nueva direccion
                                  encontre=1;
                                  limpio=1;
                               } else {
                                  error=1;
                               }
                            }
                         }
                      } else {  // encontre lugar vacio
                         encontre=1;
                         limpio=1;
                      }
                   } else {
                      error=1;
                   }
                } // (!encontre && !error)
                if ( !error ) {
                   if ( limpio ) { 
                      limpiar(&regf);
                      regf.cod = pk;
                   } else {
                      printf("El Cliente ya existe!\n");
                      mostrar(&regf);
                   }
                   regf.est = OCUPADO;
                   tipear(&regf);
                   if ( validar(&regf) ) {
                      if (preguntarSN("Cambia este Cliente?")) {
                         grabar(fp,dir,&regf);
                      }
                   }
                } // !error
             } else {
                printf("%d es un Cod.Cliente no valido (%d - %d)\n",pk,MINPK,MAXPK);
             }
          }
      } while(pk != -1);
      fclose(fp);
   } else {
      printf("%s no pudo ser abierto\n",ARCHIVO);
   }
}

void borrar() {
   FILE *fp = fopen(ARCHIVO,"r+");
   if ( fp != NULL ) {
      rg regf,regf2;
      int dir,antdir,pk;
      unsigned char encontre,error;
      do {
          pk = getIntP("Cod.Cliente (-1 para salir): ");
          if ( pk != -1 ) {
             if ( pk >= MINPK && pk <= MAXPK ) {
                dir = hashing(pk);
                antdir = VACIO;
                encontre=0;error=0;
                while (!encontre && !error) {
                   if ( leer(fp,dir,&regf) ) {
                      if ( regf.est == OCUPADO ) {
                         if ( regf.cod == pk ) { // encontre el cliente buscado
                            encontre=1;
                         } else { // maldicion! hay colision
                            if ( regf.ovr != VACIO ) { // ok, debo recorrer lista de overflow
                               antdir = dir;
                               dir = regf.ovr;
                            } else { // clave sinonima aun no dada de alta
                               encontre=1;
                               printf("El Cliente %d no ha sido creado aun!\n",pk);
                            }
                         }
                      } else {  // encontre lugar vacio
                         encontre=1;
                         printf("El Cliente %d no existe!\n",pk);
                      }
                   } else {
                      error=1;
                   }
                } // (!encontre && !error)
                if ( !error ) {
                   if ( regf.est == OCUPADO && regf.cod == pk ) {
                      mostrar(&regf);
                      if (preguntarSN("Borra este Cliente?")) {
                         // borro este registro
                         regf.est = LIBRE;
                         grabar(fp,dir,&regf);
                         if ( antdir != VACIO ) { // hay otro registro que apunta a este -> no era el primero de la lista
                            // leo anterior
                            leer(fp,antdir,&regf2);
                            if ( regf.ovr != VACIO ) {  // este registro apunta a otro -> esta en medio de la lista
                               // desconecto el actual registro, haciendo que el anterior apunte al siguiente
                               regf2.ovr = regf.ovr;
                            } else { // este registro no apunta a otro -> era el ultimo de la lista
                               // desconecto el actual registro, haciendo que el anterior no apunte mas a este
                               regf2.ovr = VACIO;
                            }
                            // grabo anterior
                            grabar(fp,antdir,&regf2);
                         } else {  // el registro es el primero de la lista, pero puede estar apuntando a otro
                            if ( regf.ovr != VACIO ) { // pongo al siguiente de la lista como primero y borro el siguiente
                               // leo el siguiente
                               leer(fp,regf.ovr,&regf2);
                               // grabo el siguiente como actual
                               grabar(fp,dir,&regf2);                              
                               // borro el siguiente
                               regf2.est = LIBRE;
                               grabar(fp,regf.ovr,&regf2);                              
                            }
                         }
                      }
                   }
                } // !error
             } else {
                printf("%d es un Cod.Cliente no valido (%d - %d)\n",pk,MINPK,MAXPK);
             }
          }
      } while(pk != -1);
      fclose(fp);
   } else {
      printf("%s no pudo ser abierto\n",ARCHIVO);
   }
}

void listar() {
   FILE *fp = fopen(ARCHIVO,"r");
   if ( fp != NULL ) {
      rg reg;
      int pos=0,upos = nuevoreg(fp);
      printf("   # Cod.  Apellido                       Nombre                       #Ap\n");
      while ( pos < upos && leer(fp,pos,&reg) ) {
         if ( reg.est == OCUPADO ) {
            printf("%4d %4d %30s %30s %4d\n",pos,reg.cod,reg.ape,reg.nom,reg.ovr);
         }
         pos++;
      }
      fclose(fp);
   } else {
      printf("%s no pudo ser abierto\n",ARCHIVO);
   }
}

void organizar() {
   FILE *fp = fopen(ARCHIVO,"r");
   if ( fp != NULL ) {
      FILE *fpt = fopen(ARCHIVOTMP,"w");
      if ( fpt != NULL ) {
         rg regf;
         int dir,tdir,n,nvadir;
         // copio todo el arreglo
         for(n=0;n<MAXREG;n++) {
            if ( leer(fp,n,&regf) ) grabar(fpt,n,&regf);
         }
         fclose(fpt);
         fpt = fopen(ARCHIVOTMP,"r+");
         // recalculo punteros
         for(n=0;n<MAXREG;n++) {
            if ( leer(fp,n,&regf) ) {
               tdir=n;dir=n;
               // recorro lista (si es que existe)
               while ( regf.ovr != VACIO ) {
                  nvadir = nuevoreg(fpt);
                  dir = regf.ovr;
                  regf.ovr = nvadir;
                  // cambio puntero a nueva direccion
                  grabar(fpt,tdir,&regf);

                  // leer siguiente nodo de la lista
                  leer(fp,dir,&regf);
 
                  // grabo nodo en area de desborde
                  grabar(fpt,nvadir,&regf);

                  tdir = nvadir;
               }
            }
         }
         fclose(fpt);
         printf("El archivo %s se encuentra reorganizado,\nchequearlo y reemplazarlo por el archivo %s\nque no se encuentra reorganizado\n",ARCHIVOTMP,ARCHIVO);
      } else {
         printf("%s no pudo ser abierto\n",ARCHIVOTMP);
      }
      fclose(fp);
   } else {
      printf("%s no pudo ser abierto\n",ARCHIVO);
   }
}

void sinonimas() {
   int pk=0,dir=0,n;
   do {
      pk = getIntP("Cod.Cliente (-1 para salir): ");
      if ( pk != -1) {
         dir=hashing(pk);
         for(n=0;n<=9999;n++) {
            if ( n != pk ) if ( dir == hashing(n) ) printf("Cod.Cliente %d es clave sinonima de %d\n",pk,n);
         }
      }
   } while (pk != -1);
}

/* funciones de bajo nivel */

void mostrar(rg *reg) {
   printf("Codigo: %d\n",reg->cod);
   printf("Apellido: %s\n",reg->ape);
   printf("Nombre: %s\n",reg->nom);
}

void tipear(rg *reg) {
   char *ap = getStringP("Apellido: ");
   char *no = getStringP("Nombre: ");
   strcpy(reg->ape,ap);
   strcpy(reg->nom,no);
   free(ap);
   free(no);
}

int validar(rg *reg) {
   int ret=1;
   if ( !(reg->cod >= MINPK && reg->cod <= MAXPK) ) { 
      printf("Codigo debe ser un numero entre %d y %d\n",MINPK,MAXPK);
      ret=0;
   } else if ( reg->ape == NULL ) {
      printf("Debe ingresar Apellido\n");
      ret=0;
   } else if ( strlen(reg->ape) == 0 ) {
      printf("El Apellido no puede estar vacio\n");
      ret=0;
   } else if ( reg->nom == NULL ) {
      printf("Debe ingresar Nombre\n");
      ret=0;
   } else if ( strlen(reg->nom) == 0 ) {
      printf("El Nombre no puede estar vacio\n");
      ret=0;
   }
   return ret;
}

void rellenar(rg *reg) {
   int n,largo1=sizeof(reg->ape),largo2=sizeof(reg->nom);
   for(n=strlen(reg->ape);n<largo1;n++) reg->ape[n] = ' ';
   for(n=strlen(reg->nom);n<largo2;n++) reg->nom[n] = ' ';
}

void limpiar(rg *reg) {
   int n,largo1=sizeof(reg->ape),largo2=sizeof(reg->nom);
   reg->est = LIBRE;
   reg->cod = VACIO;
   reg->ovr = VACIO;
   for(n=0;n<largo1;n++) reg->ape[n] = ' ';
   for(n=0;n<largo2;n++) reg->nom[n] = ' ';
}

int leer(FILE *fp,int pos,rg *reg) {
   char buf[REGLEN];
   if ( apuntar(fp,(long) pos*(REGLEN)) ) {
      if ( fread(buf,1,REGLEN,fp) == REGLEN) {
         parseReg(reg,buf);
         return 1;
      } else {
         printf("Error al intentar leer registro %d en %s\n",pos,ARCHIVO);
      }
   }
   return 0;
}

int grabar(FILE *fp,int pos,rg *reg) {
   char buf[REGLEN];
   if ( apuntar(fp,(long) pos*(REGLEN)) ) {
      rellenar(reg);
      parseBuf(buf,reg);
      if ( fwrite(buf,1,REGLEN,fp) == REGLEN ) {
         return 1;
      } else {
         printf("Error al intentar grabar registro %d en %s\n",pos,ARCHIVO);
      }
   } 
   return 0;
}

int apuntar(FILE *fp,long pos) {
   if ( ftell(fp) != pos ) {
      if ( fseek(fp,pos,SEEK_SET) == 0 ) {
         return 1;
      } else {
         printf("Error al intentar posicionarse en %s\n",ARCHIVO);
         return 0;
      }
   } else return 1;
}

void parseBuf(char *buf,rg *reg) {
   int i=0,w;
   char *p;
   buf[i++] = reg->est;
   for(w=0;w<sizeof(reg->ape);w++,i++) buf[i] = reg->ape[w];
   for(w=0;w<sizeof(reg->nom);w++,i++) buf[i] = reg->nom[w];
   p = buf+i;
   w = reg->cod;
   memcpy(p,&w,sizeof(int));
   p = buf+i+sizeof(int);
   w = reg->ovr;
   memcpy(p,&w,sizeof(int));
//   printf("pf(): ");for(i=0;i<REGLEN;i++) printf("%c",buf[i]);printf("\n");
}

void parseReg(rg *reg,char *buf) {
   int i=0,w;
   char *p;
   reg->est = buf[i++];
   for(w=0;w<sizeof(reg->ape);w++,i++) reg->ape[w] = buf[i];reg->ape[w-1]='\0';
   for(w=0;w<sizeof(reg->nom);w++,i++) reg->nom[w] = buf[i];reg->nom[w-1]='\0';
   p = buf+i;
   memcpy(&w,p,sizeof(int));
   reg->cod = w;
   p = buf+i+sizeof(int);
   memcpy(&w,p,sizeof(int));
   reg->ovr = w;
//   printf("pr(): %c %s %s %d %d\n",reg->est,reg->ape,reg->nom,reg->cod,reg->ovr);
}

int hashing(int pk) {
   return (pk % MAXREG);
}

// devuelve la direccion relativa de un nuevo registro a ser agregado al final del archivo
int nuevoreg(FILE *fp) {  
   fseek(fp,0L,SEEK_END);
   return (int) (ftell(fp)/REGLEN);
}

